import os
import shutil
import datetime
import json
import time
from nwg_displays.tools import (
    hyprctl,
    niri_msg,
    niri_reload_config,
    save_list_to_text_file,
    save_kdl_output,
    ensure_niri_config_include,
    load_text_file,
    inactive_output_description,
    load_json,
    save_json,
)
from nwg_displays.wallpaper_manager import WallpaperManager
from nwg_displays.tools import get_config

class SettingsApplier:
    @staticmethod
    def apply_from_json(profile_data, outputs_path, config_dir, profile_name):
        """Applies configuration based on a Profile JSON file."""
        SettingsApplier._save_current_state_to_previous_profile(config_dir)

        displays = profile_data["displays"]
        config = profile_data["config"]
        use_desc = config.get("use-desc", False)

        if os.getenv("NIRI_SOCKET"):
            SettingsApplier._apply_niri_json(
                displays, use_desc, outputs_path, profile_data
            )

        elif os.getenv("HYPRLAND_INSTANCE_SIGNATURE"):
            SettingsApplier._apply_hyprland_json(
                displays, use_desc, outputs_path, profile_data
            )

        elif os.getenv("SWAYSOCK"):
            SettingsApplier._apply_sway_json(displays, use_desc)

        SettingsApplier._set_active_profile(config_dir, profile_name)

    @staticmethod
    def _apply_hyprland_json(displays, use_desc, outputs_path, profile_data):
        transforms = {
            "normal": 0,
            "90": 1,
            "180": 2,
            "270": 3,
            "flipped": 4,
            "flipped-90": 5,
            "flipped-180": 6,
            "flipped-270": 7,
        }

        print(f"[Profile] Applying {len(displays)} displays for Hyprland...")
        lines = [SettingsApplier._get_header("Profile Loader")]

        for d in displays:
            if not use_desc:
                name = d["name"]
            else:
                desc_safe = d["description"].replace("#", "##")
                name = f"desc:{desc_safe}"

            if not d["active"]:
                lines.append(f"monitor={name},disable")
                hyprctl(f"dispatch dpms off {d['name']}")
                continue

            line = "monitor={},{}x{}@{},{}x{},{}".format(
                name,
                d["physical_width"],
                d["physical_height"],
                d["refresh"],
                d["x"],
                d["y"],
                d["scale"],
            )

            if d.get("mirror"):
                line += ",mirror,{}".format(d["mirror"])

            if d.get("ten_bit"):
                line += ",bitdepth,10"

            lines.append(line)

            if d["transform"] != "normal":
                t_code = transforms.get(d["transform"], 0)
                lines.append(f"monitor={name},transform,{t_code}")

            cmd = "on" if d["dpms"] else "off"
            hyprctl(f"dispatch dpms {cmd} {d['name']}")

        save_list_to_text_file(lines, outputs_path)
        hyprctl("reload")

        config, config_file = get_config()

        if "wallpapers" in profile_data and config.get(
            "profile-bound-wallpapers", True
        ):
            print("[Profile] Applying wallpapers...")
            time.sleep(1)
            WallpaperManager.apply_wallpapers(profile_data["wallpapers"])

    @staticmethod
    def _apply_niri_json(displays, use_desc, outputs_path, profile_data):
        """Apply niri configuration by writing monitor.kdl file"""
        print(f"[Profile] Applying {len(displays)} displays for niri...")
        
        kdl_data = []
        for d in displays:
            name = d["name"]
            
            display_config = {
                "name": name,
                "active": d["active"],
                "physical_width": d["physical_width"],
                "physical_height": d["physical_height"],
                "refresh": d["refresh"],
                "x": d["x"],
                "y": d["y"],
                "scale": d["scale"],
                "transform": d["transform"],
                "adaptive_sync": d.get("adaptive_sync", False)
            }
            kdl_data.append(display_config)
        
        # Save to monitor.kdl
        save_kdl_output(kdl_data, outputs_path)
        
        # Ensure config.kdl includes monitor.kdl
        niri_config_dir = os.path.dirname(outputs_path)
        ensure_niri_config_include(niri_config_dir, outputs_path)
        
        # Reload niri configuration
        niri_reload_config()
        
        config, config_file = get_config()
        
        if "wallpapers" in profile_data and config.get(
            "profile-bound-wallpapers", True
        ):
            print("[Profile] Applying wallpapers...")
            import time
            time.sleep(1)
            WallpaperManager.apply_wallpapers(profile_data["wallpapers"])

    @staticmethod
    def _apply_sway_json(displays, use_desc):
        from i3ipc import Connection

        cmds = []
        for d in displays:
            name = d["description"] if use_desc else d["name"]
            if not d["active"]:
                cmds.append(f'output "{name}" disable')
                continue

            cmd = 'output "{}"'.format(name)
            custom = "--custom" if d["custom_mode"] else ""
            cmd += " mode {} {}x{}@{}Hz".format(
                custom, d["physical_width"], d["physical_height"], d["refresh"]
            )
            cmd += " pos {} {}".format(d["x"], d["y"])
            cmd += " transform {}".format(d["transform"])
            cmd += " scale {}".format(d["scale"])

            if d.get("scale_filter"):
                cmd += " scale_filter {}".format(d["scale_filter"])

            a_s = "on" if d["adaptive_sync"] else "off"
            cmd += " adaptive_sync {}".format(a_s)

            dpms = "on" if d["dpms"] else "off"
            cmd += " dpms {}".format(dpms)
            cmds.append(cmd)

        i3 = Connection()
        for cmd in cmds:
            i3.command(cmd)

    @staticmethod
    def apply_from_gui(
        display_buttons,
        outputs_activity,
        outputs_path,
        use_desc=False,
        create_confirm_win_callback=None,
        config_dir=None,
        profile_name=None,
    ):
        """
        Applies configuration based on GUI buttons state.
        Refactored from original 'apply_settings'.
        """
        if config_dir:
            SettingsApplier._save_current_state_to_previous_profile(config_dir)

        if os.getenv("NIRI_SOCKET"):
            print(f"[DEBUG] Applying niri config to {outputs_path}")
            SettingsApplier._apply_niri_gui(
                display_buttons,
                outputs_activity,
                outputs_path,
                use_desc,
                create_confirm_win_callback,
                config_dir,
                profile_name,
            )

        elif os.getenv("SWAYSOCK"):
            SettingsApplier._apply_sway_gui(
                display_buttons,
                outputs_activity,
                outputs_path,
                use_desc,
                create_confirm_win_callback,
                config_dir,
                profile_name,
            )

        elif os.getenv("HYPRLAND_INSTANCE_SIGNATURE"):
            SettingsApplier._apply_hyprland_gui(
                display_buttons,
                outputs_activity,
                outputs_path,
                use_desc,
                create_confirm_win_callback,
                config_dir,
                profile_name,
            )
        else:
            print("[Error] No compositor detected (Sway/Hyprland/Niri)")

        if config_dir and profile_name:
            SettingsApplier._set_active_profile(config_dir, profile_name)

    @staticmethod
    def _apply_sway_gui(
        display_buttons,
        outputs_activity,
        outputs_path,
        use_desc,
        create_confirm_win_callback,
        config_dir=None,
        profile_name=None,
    ):
        from i3ipc import Connection

        lines = [SettingsApplier._get_header()]
        cmds = []
        db_names = []

        for db in display_buttons:
            name = db.name if not use_desc else db.description
            db_names.append(name)

            lines.append('output "%s" {' % name)
            cmd = 'output "{}"'.format(name)

            custom_mode_str = "--custom" if db.custom_mode else ""
            lines.append(
                "    mode {} {}x{}@{}Hz".format(
                    custom_mode_str,
                    db.physical_width,
                    db.physical_height,
                    db.refresh,
                )
            )
            cmd += " mode {} {}x{}@{}Hz".format(
                custom_mode_str, db.physical_width, db.physical_height, db.refresh
            )

            lines.append("    pos {} {}".format(db.x, db.y))
            cmd += " pos {} {}".format(db.x, db.y)

            lines.append("    transform {}".format(db.transform))
            cmd += " transform {}".format(db.transform)

            lines.append("    scale {}".format(db.scale))
            cmd += " scale {}".format(db.scale)

            lines.append("    scale_filter {}".format(db.scale_filter))
            cmd += " scale_filter {}".format(db.scale_filter)

            a_s = "on" if db.adaptive_sync else "off"
            lines.append("    adaptive_sync {}".format(a_s))
            cmd += " adaptive_sync {}".format(a_s)

            dpms = "on" if db.dpms else "off"
            lines.append("    dpms {}".format(dpms))
            cmd += " dpms {}".format(dpms)

            lines.append("}")
            cmds.append(cmd)

        if not use_desc:
            for key in outputs_activity:
                if key not in db_names:
                    lines.append('output "{}" disable'.format(key))
                    cmds.append('output "{}" disable'.format(key))
        else:
            for key in outputs_activity:
                desc = inactive_output_description(key)
                if desc not in db_names:
                    lines.append('output "{}" disable'.format(desc))
                    cmds.append('output "{}" disable'.format(desc))

        if os.path.isfile(outputs_path):
            backup = load_text_file(outputs_path).splitlines()
        else:
            backup = []

        save_list_to_text_file(lines, outputs_path)

        i3 = Connection()
        for cmd in cmds:
            i3.command(cmd)

        if create_confirm_win_callback:
            create_confirm_win_callback(backup, outputs_path, config_dir, profile_name)

    @staticmethod
    def _apply_niri_gui(
        display_buttons,
        outputs_activity,
        outputs_path,
        use_desc,
        create_confirm_win_callback,
        config_dir=None,
        profile_name=None,
    ):
        """Apply niri configuration from GUI by writing monitor.kdl file"""
        print(f"[niri] Applying {len(display_buttons)} displays...")
        
        # Save backup BEFORE applying new settings
        backup_path = outputs_path + ".bak"
        if os.path.isfile(outputs_path):
            shutil.copy2(outputs_path, backup_path)
            print(f"[niri] Backup saved to {backup_path}")
        else:
            backup_path = None
        
        kdl_data = []
        for db in display_buttons:
            display_config = {
                "name": db.name,
                "active": db.name not in outputs_activity or outputs_activity.get(db.name, True),
                "physical_width": db.physical_width,
                "physical_height": db.physical_height,
                "refresh": db.refresh,
                "x": db.x,
                "y": db.y,
                "scale": db.scale,
                "transform": db.transform,
                "adaptive_sync": db.adaptive_sync
            }
            kdl_data.append(display_config)
        
        # Save to monitor.kdl in KDL format
        save_kdl_output(kdl_data, outputs_path)
        
        # Ensure config.kdl includes monitor.kdl
        niri_config_dir = os.path.dirname(outputs_path)
        ensure_niri_config_include(niri_config_dir, outputs_path)
        
        # Reload niri configuration
        niri_reload_config()
        
        # Pass backup file path and current file path to confirm window
        if create_confirm_win_callback:
            create_confirm_win_callback(backup_path, outputs_path, config_dir, profile_name)

    @staticmethod
    def _apply_hyprland_gui(
        display_buttons,
        outputs_activity,
        outputs_path,
        use_desc,
        create_confirm_win_callback,
        config_dir=None,
        profile_name=None,
    ):
        transforms = {
            "normal": 0,
            "90": 1,
            "180": 2,
            "270": 3,
            "flipped": 4,
            "flipped-90": 5,
            "flipped-180": 6,
            "flipped-270": 7,
        }
        lines = [SettingsApplier._get_header()]

        for db in display_buttons:
            name = (
                db.name
                if not use_desc
                else "desc:{}".format(db.description.replace("#", "##"))
            )

            if db.name in outputs_activity and not outputs_activity[db.name]:
                lines.append("monitor={},disable".format(name))
                hyprctl(f"dispatch dpms off {db.name}")
                continue

            # Format: monitor=name,resolution@refresh,position,scale
            line = "monitor={},{}x{}@{},{}x{},{}".format(
                name,
                db.physical_width,
                db.physical_height,
                db.refresh,
                db.x,
                db.y,
                db.scale,
            )
            if db.mirror:
                line += ",mirror,{}".format(db.mirror)
            if db.ten_bit:
                line += ",bitdepth,10"

            lines.append(line)
            if db.transform != "normal":
                lines.append(
                    "monitor={},transform,{}".format(name, transforms[db.transform])
                )

            cmd = "on" if db.dpms else "off"
            hyprctl(f"dispatch dpms {cmd} {db.name}")

        backup = []
        if os.path.isfile(outputs_path):
            backup = load_text_file(outputs_path).splitlines()
        save_list_to_text_file(lines, outputs_path)
        hyprctl("reload")

        if create_confirm_win_callback:
            create_confirm_win_callback(backup, outputs_path, config_dir, profile_name)

    @staticmethod
    def _get_header(source="nwg-displays"):
        now = datetime.datetime.now()
        return "# Generated by {} on {} at {}. Do not edit manually.\n".format(
            source,
            datetime.datetime.strftime(now, "%Y-%m-%d"),
            datetime.datetime.strftime(now, "%H:%M:%S"),
        )

    @staticmethod
    def _save_current_state_to_previous_profile(config_dir):
        """
        Reads the last active profile name, gets current wallpapers,
        and updates that profile's JSON file.
        """
        state_file = os.path.join(config_dir, "active_profile.json")

        if not os.path.isfile(state_file):
            return

        try:
            data = load_json(state_file)
            last_profile_name = data.get("active_profile") if data else None

            if not last_profile_name:
                return

            config, _ = get_config()
            if not config.get("profile-bound-wallpapers", True):
                return

            prev_profile_path = os.path.join(
                config_dir, "profiles", f"{last_profile_name}.json"
            )

            if not os.path.isfile(prev_profile_path):
                print(
                    f"[Warning] Previous profile '{last_profile_name}' file not found. Skipping save."
                )
                return

            current_walls = WallpaperManager.get_current_wallpapers()
            if not current_walls:
                return

            with open(prev_profile_path, "r") as f:
                data = json.load(f)

            if "wallpapers" not in data:
                data["wallpapers"] = {}

            data["wallpapers"].update(current_walls)

            with open(prev_profile_path, "w") as f:
                json.dump(data, f, indent=2)

            print(f"[Persistence] Saved current wallpapers to '{last_profile_name}'")

        except Exception as e:
            print(f"[Error] Failed to save previous state: {e}")

    @staticmethod
    def _set_active_profile(config_dir, profile_name):
        state_file = os.path.join(config_dir, "active_profile.json")
        try:
            save_json({"active_profile": profile_name}, state_file)
        except Exception as e:
            print(f"[Error] Failed to set active profile: {e}")



import os
import sys
import json
import argparse
from nwg_displays.settings_applier import SettingsApplier
from nwg_displays.tools import get_config_dir


def main():
    parser = argparse.ArgumentParser(description="Load nwg-displays profile directly.")
    parser.add_argument(
        "-p", "--profile", type=str, required=True, help="Name of the profile to load"
    )
    parser.add_argument(
        "-c", "--config", type=str, help="Path to monitors.conf override"
    )

    args = parser.parse_args()

    config_dir = get_config_dir()
    profile_path = os.path.join(config_dir, "profiles", f"{args.profile}.json")

    if args.config:
        outputs_path = args.config
    else:
        outputs_path = os.path.join(
            os.path.expanduser("~/.config/hypr/"), "monitors.conf"
        )

    if not os.path.isfile(profile_path):
        print(f"[Error] Profile file not found at {profile_path}")
        sys.exit(1)

    try:
        with open(profile_path, "r") as f:
            profile_data = json.load(f)

        print(f"[Profile] Loading '{args.profile}'")

        SettingsApplier.apply_from_json(
            profile_data, outputs_path, config_dir, args.profile
        )

        print("[Profile] Done.")

    except json.JSONDecodeError:
        print(f"[Error] Failed to parse {profile_path}. Invalid JSON.")
        sys.exit(1)
    except Exception as e:
        print(f"[Error] An unexpected error occurred: {e}")
        sys.exit(1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python

"""
Output management utility for sway Wayland compositor, inspired by wdisplays and wlay
Project: https://github.com/nwg-piotr/nwg-displays
Author's email: nwg.piotr@gmail.com
Copyright (c) 2022-2024 Piotr Miller & Contributors
License: MIT
Depends on: 'python-i3ipc' 'gtk-layer-shell'

All the code below was built around this glorious snippet:
https://gist.github.com/KurtJacobson/57679e5036dc78e6a7a3ba5e0155dad1
Thank you, Kurt Jacobson!
"""

import argparse
import shutil
import stat
import sys
import threading
import gi
from nwg_displays.settings_applier import SettingsApplier
from nwg_displays.wallpaper_manager import WallpaperManager


gi.require_version("Gtk", "3.0")
try:
    gi.require_version("GtkLayerShell", "0.1")
except ValueError:
    raise RuntimeError(
        "\n\n"
        + "If you haven't installed GTK Layer Shell, you need to point Python to the\n"
        + "library by setting GI_TYPELIB_PATH and LD_LIBRARY_PATH to <build-dir>/src/.\n"
        + "For example you might need to run:\n\n"
        + "GI_TYPELIB_PATH=build/src LD_LIBRARY_PATH=build/src python3 "
        + " ".join(sys.argv)
    )

from gi.repository import Gtk, GLib, GtkLayerShell

from nwg_displays.tools import *
from nwg_displays.profiles import ProfileManager
from nwg_displays.__about__ import __version__

dir_name = os.path.dirname(__file__)
sway = os.getenv("SWAYSOCK") is not None
hypr = os.getenv("HYPRLAND_INSTANCE_SIGNATURE") is not None
niri = os.getenv("NIRI_SOCKET") is not None

config_dir = os.path.join(get_config_home(), "nwg-displays")
# This was done by mistake, and the config file need to be migrated to the proper path
old_config_dir = os.path.join(get_config_home(), "nwg-outputs")

sway_config_dir = os.path.join(get_config_home(), "sway")
if sway and not os.path.isdir(sway_config_dir):
    print(
        "[Warning] Couldn't find sway config directory '{}'".format(sway_config_dir),
        file=sys.stderr,
    )
    sys.exit(1)

hypr_config_dir = os.path.join(get_config_home(), "hypr")
if hypr and not os.path.isdir(hypr_config_dir):
    print(
        "[Warning] Couldn't find Hyprland config directory '{}'".format(
            hypr_config_dir
        ),
        file=sys.stderr,
    )
    sys.exit(1)

niri_config_dir = os.path.join(get_config_home(), "niri")
if niri and not os.path.isdir(niri_config_dir):
    print(
        "[Warning] Couldn't find niri config directory '{}'".format(niri_config_dir),
        file=sys.stderr,
    )
    sys.exit(1)

# Create empty files if not found
if sway:
    for name in ["outputs", "workspaces"]:
        create_empty_file(os.path.join(sway_config_dir, name))
elif hypr:
    for name in ["monitors.conf", "workspaces.conf"]:
        create_empty_file(os.path.join(hypr_config_dir, name))
elif niri:
    for name in ["monitor.kdl"]:
        create_empty_file(os.path.join(niri_config_dir, name))
else:
    eprint("[Error] Neither sway, Hyprland nor niri detected, terminating")
    sys.exit(1)

config = {}
outputs_path = ""
num_ws = 0

"""
i3.get_outputs() does not return some output attributes, especially when connected via hdmi.
i3.get_tree() on the other hand does not return inactive outputs. So we'll list attributes with .get_tree(),
and the add inactive outputs, if any, from what we detect with .get_outputs()
"""
outputs = (
    {}
)  # Active outputs, listed from the sway tree; stores name and all attributes.
outputs_activity = {}  # Just a dictionary "name": is_active - from get_outputs()
workspaces = {}  # "workspace_num": "display_name"

display_buttons = []
selected_output_button = None
profile_manager = None

# Glade form fields
form_name = None
form_description = None
form_dpms = None
form_adaptive_sync = None
form_custom_mode = None
form_view_scale = None
form_use_desc = None
form_x = None
form_y = None
form_width = None
form_height = None
form_scale = None
form_scale_filter = None
form_refresh = None
form_modes = None
form_transform = None
form_wrapper_box = None
form_workspaces = None
form_close = None
form_apply = None
form_version = None
form_mirror = None
form_ten_bit = None
form_profile_wallpapers = None

dialog_win = None
confirm_win = None
src_tag = 0
counter = 0

"""
We need to rebuild the modes GtkComboBoxText on each DisplayButton click. Unfortunately appending an item fires the
"change" event every time (and we have no "value-changed" event here). Setting `on_mode_changed_silent` True will 
prevent the `on_mode_changed` function from working.
"""
on_mode_changed_silent = False

# Value from config adjusted to current view scale
snap_threshold_scaled = None

fixed = Gtk.Fixed()

SENSITIVITY = 1

EvMask = Gdk.EventMask.BUTTON_PRESS_MASK | Gdk.EventMask.BUTTON1_MOTION_MASK

offset_x = 0
offset_y = 0
px = 0
py = 0
max_x = 0
max_y = 0

voc = {}


def load_vocabulary():
    global voc
    # basic vocabulary (for en_US)
    voc = load_json(os.path.join(dir_name, "langs", "en_US.json"))
    if not voc:
        eprint("Failed loading vocabulary, terminating")
        sys.exit(1)

    shell_data = load_shell_data()

    lang = os.getenv("LANG")
    if lang is None:
        lang = "en_US"
    else:
        lang = (
            lang.split(".")[0]
            if not shell_data["interface-locale"]
            else shell_data["interface-locale"]
        )

    # translate if translation available
    if lang != "en_US":
        loc_file = os.path.join(dir_name, "langs", "{}.json".format(lang))
        if os.path.isfile(loc_file):
            # localized vocabulary
            loc = load_json(loc_file)
            if not loc:
                eprint("Failed loading translation into '{}'".format(lang))
            else:
                for key in loc:
                    voc[key] = loc[key]


def on_button_press_event(widget, event):
    if widget != selected_output_button:
        widget.indicator.show_up()

    if event.button == 1:
        for db in display_buttons:
            if db.name == widget.name:
                db.select()
            else:
                db.unselect()

        p = widget.get_parent()
        # offset == distance of parent widget from edge of screen ...
        global offset_x, offset_y
        offset_x, offset_y = p.get_window().get_position()
        # plus distance from pointer to edge of widget
        offset_x += event.x
        offset_y += event.y
        # max_x, max_y both relative to the parent
        # note that we're rounding down now so that these max values don't get
        # rounded upward later and push the widget off the edge of its parent.
        global max_x, max_y
        max_x = round_down_to_multiple(
            p.get_allocation().width - widget.get_allocation().width, SENSITIVITY
        )
        max_y = round_down_to_multiple(
            p.get_allocation().height - widget.get_allocation().height, SENSITIVITY
        )

        update_form_from_widget(widget)


def on_motion_notify_event(widget, event):
    # x_root,x_root relative to screen
    # x,y relative to parent (fixed widget)
    # px,py stores previous values of x,y

    global px, py

    # get starting values for x,y
    x = event.x_root - offset_x
    y = event.y_root - offset_y
    # make sure the potential coordinates x,y:
    #   1) will not push any part of the widget outside of its parent container
    #   2) is a multiple of SENSITIVITY
    x = round_to_nearest_multiple(max_val(min_val(x, max_x), 0), SENSITIVITY)
    y = round_to_nearest_multiple(max_val(min_val(y, max_y), 0), SENSITIVITY)

    if x != px or y != py:
        px = x
        py = y
        snap_x, snap_y = [0], [0]
        # Collect snap lines from other displays
        for db in display_buttons:
            if db.name == widget.name:
                continue

            val = db.x * config["view-scale"]
            if val not in snap_x:
                snap_x.append(val)

            val = (db.x + db.logical_width) * config["view-scale"]
            if val not in snap_x:
                snap_x.append(val)

            val = db.y * config["view-scale"]
            if val not in snap_y:
                snap_y.append(val)

            val = (db.y + db.logical_height) * config["view-scale"]
            if val not in snap_y:
                snap_y.append(val)

        snap_h, snap_v = None, None
        # Find nearest horizontal snap line
        for value in snap_x:
            if abs(x - value) < snap_threshold_scaled:
                snap_h = value
                break

        for value in snap_x:
            w = widget.logical_width * config["view-scale"]
            if abs(w + x - value) < snap_threshold_scaled:
                snap_h = value - w
                break

        # Find nearest vertical snap line
        for value in snap_y:
            if abs(y - value) < snap_threshold_scaled:
                snap_v = value
                break

        for value in snap_y:
            h = widget.logical_height * config["view-scale"]
            if abs(h + y - value) < snap_threshold_scaled:
                snap_v = value - h
                break

        # Just in case ;)
        if snap_h and snap_h < 0:
            snap_h = 0

        if snap_v and snap_v < 0:
            snap_v = 0

        if snap_h is None and snap_v is None:
            fixed.move(widget, x, y)
            widget.x = round(x / config["view-scale"])
            widget.y = round(y / config["view-scale"])
        else:

            if snap_h is not None and snap_v is not None:
                fixed.move(widget, snap_h, snap_v)
                widget.x = round(snap_h / config["view-scale"])
                widget.y = round(snap_v / config["view-scale"])

            elif snap_h is not None:
                fixed.move(widget, snap_h, y)
                widget.x = round(snap_h / config["view-scale"])
                widget.y = round(y / config["view-scale"])

            elif snap_v is not None:
                fixed.move(widget, x, snap_v)
                widget.x = round(x / config["view-scale"])
                widget.y = round(snap_v / config["view-scale"])

    update_form_from_widget(widget)


def update_form_from_widget(widget):
    form_name.set_text(widget.name)
    if len(widget.description) > 48:
        form_description.set_text(f"{widget.description[:47]}(…)")
    else:
        form_description.set_text(widget.description)
    form_dpms.set_active(widget.dpms)
    form_adaptive_sync.set_active(widget.adaptive_sync)
    form_custom_mode.set_active(widget.custom_mode)
    form_view_scale.set_value(
        config["view-scale"]
    )  # not really from the widget, but from the global value
    if form_profile_wallpapers:
        form_profile_wallpapers.set_active(config.get("profile-bound-wallpapers", True))
    form_use_desc.set_active(config["use-desc"])
    form_x.set_value(widget.x)
    form_y.set_value(widget.y)
    form_width.set_value(widget.physical_width)
    form_height.set_value(widget.physical_height)
    form_scale.set_value(widget.scale)
    form_scale_filter.set_active_id(widget.scale_filter)
    form_refresh.set_value(widget.refresh)
    if form_ten_bit:
        form_ten_bit.set_active(widget.ten_bit)
    if form_mirror:
        form_mirror.remove_all()
        form_mirror.append("", voc["none"])
        for key in outputs:
            if key != widget.name:
                form_mirror.append(key, key)
        form_mirror.set_active_id(widget.mirror)
        form_mirror.show_all()

    global on_mode_changed_silent
    on_mode_changed_silent = True

    form_modes.remove_all()
    active = ""
    for mode in widget.modes:
        m = "{}x{}@{}Hz".format(
            mode["width"],
            mode["height"],
            mode["refresh"] / 1000,
            mode["refresh"] / 1000,
            widget.refresh,
        )
        form_modes.append(m, m)
        # This is just to set active_id

        if (
            mode["width"] == widget.physical_width
            and mode["height"] == widget.physical_height
            and mode["refresh"] / 1000 == widget.refresh
        ):
            active = m
    if active:
        form_modes.set_active_id(active)

    form_transform.set_active_id(widget.transform)

    on_mode_changed_silent = False


class DisplayButton(Gtk.Button):
    def __init__(
        self,
        name,
        description,
        x,
        y,
        physical_width,
        physical_height,
        transform,
        scale,
        scale_filter,
        refresh,
        modes,
        active,
        dpms,
        adaptive_sync_status,
        ten_bit,
        custom_mode_status,
        focused,
        monitor,
        mirror="",
    ):
        super().__init__()
        # Output properties
        self.name = name
        self.description = description
        self.x = x
        self.y = y
        self.physical_width = physical_width
        self.physical_height = physical_height
        self.transform = transform
        self.scale = scale
        self.scale_filter = scale_filter
        self.refresh = refresh
        self.modes = []
        for m in modes:
            if m not in self.modes:
                self.modes.append(m)
        # self.modes = modes
        self.active = active
        self.dpms = dpms
        self.adaptive_sync = (
            adaptive_sync_status == "enabled"
        )  # converts "enabled | disabled" to bool
        self.custom_mode = custom_mode_status
        self.focused = focused
        self.mirror = mirror
        self.ten_bit = ten_bit

        # Button properties
        self.selected = False
        self.set_can_focus(False)
        self.set_events(EvMask)
        self.connect("button_press_event", on_button_press_event)
        self.connect("motion_notify_event", on_motion_notify_event)
        self.set_always_show_image(True)
        self.set_label(self.name)

        self.rescale_transform()
        self.set_property("name", "output")

        self.indicator = Indicator(
            monitor,
            name,
            round(self.physical_width * config["view-scale"]),
            round(self.physical_height * config["view-scale"]),
            config["indicator-timeout"],
        )

        self.show()

    @property
    def logical_width(self):
        if is_rotated(self.transform):
            return self.physical_height / self.scale
        else:
            return self.physical_width / self.scale

    @property
    def logical_height(self):
        if is_rotated(self.transform):
            return self.physical_width / self.scale
        else:
            return self.physical_height / self.scale

    def select(self):
        self.selected = True
        self.set_property("name", "selected-output")
        global selected_output_button
        selected_output_button = self

    def unselect(self):
        self.selected = False
        self.set_property("name", "output")

    def rescale_transform(self):
        self.set_size_request(
            round(self.logical_width * config["view-scale"]),
            round(self.logical_height * config["view-scale"]),
        )

    def on_active_check_button_toggled(self, w):
        self.active = w.get_active()
        if not self.active:
            self.set_property("name", "inactive-output")
        else:
            if self == selected_output_button:
                self.set_property("name", "selected-output")
            else:
                self.set_property("name", "output")


def on_view_scale_changed(*args):
    config["view-scale"] = round(form_view_scale.get_value(), 2)

    global snap_threshold_scaled
    snap_threshold_scaled = round(config["snap-threshold"] * config["view-scale"] * 10)

    for b in display_buttons:
        b.rescale_transform()
        fixed.move(b, b.x * config["view-scale"], b.y * config["view-scale"])

    save_json(config, os.path.join(config_dir, "config"))


def on_profile_wallpapers_toggled(widget):
    config["profile-bound-wallpapers"] = widget.get_active()
    save_json(config, os.path.join(config_dir, "config"))


def on_transform_changed(*args):
    if selected_output_button:
        transform = form_transform.get_active_id()
        selected_output_button.transform = transform
        selected_output_button.rescale_transform()


def on_ten_bit_toggled(check_btn):
    if selected_output_button:
        selected_output_button.ten_bit = check_btn.get_active()


def on_dpms_toggled(widget):
    if selected_output_button:
        selected_output_button.dpms = widget.get_active()


def on_use_desc_toggled(widget):
    config["use-desc"] = widget.get_active()
    save_json(config, os.path.join(config_dir, "config"))


def on_adaptive_sync_toggled(widget):
    if selected_output_button:
        selected_output_button.adaptive_sync = widget.get_active()


def on_custom_mode_toggle(widget):
    if selected_output_button:
        outputs = set(config["custom-mode"])
        turned_on = widget.get_active()
        selected_output_button.custom_mode = turned_on
        if turned_on:
            outputs.add(selected_output_button.name)
        else:
            outputs.discard(selected_output_button.name)
        config["custom-mode"] = tuple(outputs)


def on_pos_x_changed(widget):
    if selected_output_button:
        selected_output_button.x = round(widget.get_value())
        fixed.move(
            selected_output_button,
            selected_output_button.x * config["view-scale"],
            selected_output_button.y * config["view-scale"],
        )


def on_pos_y_changed(widget):
    if selected_output_button:
        selected_output_button.y = round(widget.get_value())
        fixed.move(
            selected_output_button,
            selected_output_button.x * config["view-scale"],
            selected_output_button.y * config["view-scale"],
        )


def on_width_changed(widget):
    if selected_output_button:
        selected_output_button.physical_width = round(widget.get_value())
        selected_output_button.rescale_transform()


def on_height_changed(widget):
    if selected_output_button:
        selected_output_button.physical_height = round(widget.get_value())
        selected_output_button.rescale_transform()


def on_scale_changed(widget):
    if selected_output_button:
        selected_output_button.scale = widget.get_value()
        selected_output_button.rescale_transform()


def on_scale_filter_changed(widget):
    if selected_output_button:
        selected_output_button.scale_filter = widget.get_active_id()


def on_refresh_changed(widget):
    if selected_output_button:
        selected_output_button.refresh = widget.get_value()

        update_form_from_widget(selected_output_button)


def on_mode_changed(widget):
    if selected_output_button and not on_mode_changed_silent:
        mode = selected_output_button.modes[widget.get_active()]
        selected_output_button.physical_width = mode["width"]
        selected_output_button.physical_height = mode["height"]
        selected_output_button.refresh = mode["refresh"] / 1000
        selected_output_button.rescale_transform()

        update_form_from_widget(selected_output_button)


def on_mirror_selected(widget):
    if selected_output_button and widget.get_active_id() is not None:
        selected_output_button.mirror = widget.get_active_id()


def on_apply_button(widget, p_manager=None):
    if p_manager is None:
        p_manager = profile_manager

    profile_name = p_manager.current_profile if p_manager else None
    # Remove legacy storage
    if "active-profile" in config:
        del config["active-profile"]

    SettingsApplier.apply_from_gui(
        display_buttons,
        outputs_activity,
        outputs_path,
        config["use-desc"],
        create_confirm_win,
        config_dir,
        profile_name=profile_name,
    )
    # save config file
    save_json(config, os.path.join(config_dir, "config"))


def on_output_toggled(check_btn, name):
    outputs_activity[name] = check_btn.get_active()


def on_toggle_button(btn):
    i3 = Connection()
    for key in outputs_activity:
        toggle = "enable" if outputs_activity[key] else "disable"
        cmd = "output {} {}".format(key, toggle)
        i3.command(cmd)

    # If the output has just been turned back on, Gdk.Display.get_default() may need some time
    GLib.timeout_add(1000, create_display_buttons)


def create_display_buttons():
    global display_buttons, outputs
    for item in display_buttons:
        item.destroy()
    display_buttons = []

    outputs = list_outputs()
    
    if not outputs:
        eprint("[Error] No outputs detected")
        return
    
    for key in outputs:
        item = outputs[key]
        custom_mode = key in config["custom-mode"]
        b = DisplayButton(
            key,
            item["description"],
            item["x"],
            item["y"],
            round(item["physical-width"]),
            round(item["physical-height"]),
            item["transform"],
            item["scale"],
            item["scale_filter"],
            item["refresh"],
            item["modes"],
            item["active"],
            item["dpms"],
            item["adaptive_sync_status"],
            item["ten_bit"],
            custom_mode,
            item["focused"],
            item["monitor"],
            mirror=item["mirror"],
        )

        display_buttons.append(b)

        fixed.put(
            b,
            round(item["x"] * config["view-scale"]),
            round(item["y"] * config["view-scale"]),
        )

    if display_buttons:
        display_buttons[0].select()
        update_form_from_widget(display_buttons[0])


class Indicator(Gtk.Window):
    def __init__(self, monitor, name, width, height, timeout):
        super().__init__()
        self.timeout = timeout
        self.monitor = monitor
        self.set_property("name", "indicator")

        GtkLayerShell.init_for_window(self)
        GtkLayerShell.set_layer(self, GtkLayerShell.Layer.OVERLAY)
        if monitor:
            GtkLayerShell.set_monitor(self, monitor)

        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.add(box)
        label = Gtk.Label()
        box.set_property("name", "indicator-label")
        label.set_text(name)
        box.pack_start(label, True, True, 10)

        self.set_size_request(width, height)
        if self.timeout > 0:
            self.show_up(self.timeout * 2)

    def show_up(self, timeout=None):
        if self.timeout > 0 and self.monitor:
            self.show_all()
            if timeout:
                GLib.timeout_add(timeout, self.hide)
            else:
                GLib.timeout_add(self.timeout, self.hide)


def handle_keyboard(window, event):
    if event.type == Gdk.EventType.KEY_RELEASE and event.keyval == Gdk.KEY_Escape:
        window.close()


def create_workspaces_window(btn):
    global workspaces
    workspaces = load_workspaces(
        os.path.join(sway_config_dir, "workspaces"), use_desc=config["use-desc"]
    )
    old_workspaces = workspaces.copy()
    global dialog_win
    if dialog_win:
        dialog_win.destroy()
    dialog_win = Gtk.Window()
    dialog_win.set_resizable(False)
    dialog_win.set_modal(True)
    dialog_win.connect("key-release-event", handle_keyboard)
    grid = Gtk.Grid()
    for prop in ["margin_start", "margin_end", "margin_top", "margin_bottom"]:
        grid.set_property(prop, 10)
    grid.set_column_spacing(12)
    grid.set_row_spacing(12)
    dialog_win.add(grid)
    last_row = 0
    for i in range(num_ws):
        lbl = Gtk.Label()
        lbl.set_text("workspace {} output ".format(i + 1))
        grid.attach(lbl, 0, i, 1, 1)
        combo = Gtk.ComboBoxText()
        for key in outputs:
            if not config["use-desc"]:
                combo.append(key, key)
            else:
                desc = "{}".format(outputs[key]["description"])
                combo.append(desc, desc)
            if i + 1 in workspaces:
                combo.set_active_id(workspaces[i + 1])
            combo.connect("changed", on_ws_combo_changed, i + 1)
        grid.attach(combo, 1, i, 1, 1)
        last_row = i

    box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
    grid.attach(box, 0, last_row + 1, 2, 1)

    btn_apply = Gtk.Button()
    btn_apply.set_label(voc["apply"])
    if sway_config_dir:
        btn_apply.connect(
            "clicked", on_workspaces_apply_btn, dialog_win, old_workspaces
        )
    else:
        btn_apply.set_sensitive(False)
        btn_apply.set_tooltip_text("Config dir not found")
    box.pack_end(btn_apply, False, False, 0)

    btn_close = Gtk.Button()
    btn_close.set_label(voc["close"])
    btn_close.connect("clicked", close_dialog, dialog_win)
    box.pack_end(btn_close, False, False, 6)

    dialog_win.show_all()


def create_workspaces_window_hypr(btn):
    global workspaces
    workspaces = load_workspaces_hypr(
        os.path.join(hypr_config_dir, "workspaces.conf"), num_ws=num_ws
    )
    eprint("WS->Mon:", workspaces)
    old_workspaces = workspaces.copy()
    global dialog_win
    if dialog_win:
        dialog_win.destroy()
    dialog_win = Gtk.Window()
    dialog_win.set_resizable(False)
    dialog_win.set_modal(True)
    dialog_win.connect("key-release-event", handle_keyboard)
    grid = Gtk.Grid()
    for prop in ["margin_start", "margin_end", "margin_top", "margin_bottom"]:
        grid.set_property(prop, 10)
    grid.set_column_spacing(12)
    grid.set_row_spacing(6)
    dialog_win.add(grid)
    last_row = 0
    for i in range(num_ws):
        lbl = Gtk.Label()
        if config["use-desc"]:
            lbl.set_markup(
                "Workspace rule: <b>workspace={},monitor:desc:</b>".format(i + 1)
            )
        else:
            lbl.set_markup("Workspace rule: <b>workspace={},monitor:</b>".format(i + 1))
        lbl.set_property("halign", Gtk.Align.END)
        grid.attach(lbl, 0, i, 1, 1)
        combo = Gtk.ComboBoxText()
        for key in outputs:
            if not config["use-desc"]:
                combo.append(key, key)
            else:
                desc = "{}".format(outputs[key]["description"])
                combo.append(desc, desc)
            if i + 1 in workspaces:
                combo.set_active_id(workspaces[i + 1])
            combo.connect("changed", on_ws_combo_changed, i + 1)

        grid.attach(combo, 1, i, 1, 1)

        last_row += 1

    box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
    grid.attach(box, 0, last_row + 1, 2, 1)

    btn_apply = Gtk.Button()
    btn_apply.set_label(voc["apply"])
    if hypr_config_dir:
        btn_apply.connect(
            "clicked", on_workspaces_apply_btn_hypr, dialog_win, old_workspaces
        )
    else:
        btn_apply.set_sensitive(False)
        btn_apply.set_tooltip_text("Config dir not found")
    box.pack_end(btn_apply, False, False, 0)

    btn_close = Gtk.Button()
    btn_close.set_label(voc["close"])
    btn_close.connect("clicked", close_dialog, dialog_win)
    box.pack_end(btn_close, False, False, 6)

    dialog_win.show_all()


def on_ws_combo_changed(combo, ws_num):
    workspaces[ws_num] = combo.get_active_id()


def close_dialog(w, win):
    win.close()


def on_workspaces_apply_btn(w, win, old_workspaces):
    if workspaces != old_workspaces:
        save_workspaces(
            workspaces,
            os.path.join(sway_config_dir, "workspaces"),
            use_desc=config["use-desc"],
        )
        notify("Workspaces assignment", "Restart sway for changes to take effect")

    close_dialog(w, win)


def on_workspaces_apply_btn_hypr(w, win, old_workspaces):
    if workspaces != old_workspaces:
        workspace_conf_file = workspaces_path
        text_file = open(workspace_conf_file, "w")

        now = datetime.datetime.now()
        line = (
            "# Generated by nwg-displays on {} at {}. Do not edit manually.\n".format(
                datetime.datetime.strftime(now, "%Y-%m-%d"),
                datetime.datetime.strftime(now, "%H:%M:%S"),
            )
        )
        text_file.write(line + "\n")

        monitors_with_default_workspace = []
        for ws in workspaces:
            mon = workspaces[ws]
            if not config["use-desc"]:
                line = "workspace={},monitor:{}".format(ws, mon)
            else:
                line = "workspace={},monitor:desc:{}".format(ws, mon)

            if mon not in monitors_with_default_workspace:
                line += ",default:true"
                monitors_with_default_workspace.append(mon)

            text_file.write(line + "\n")

        text_file.close()
        notify("Workspaces assignment", "Restart Hyprland for changes to take effect")

    close_dialog(w, win)


def create_confirm_win(backup, path, config_dir=None, profile_name=None):
    global confirm_win
    if confirm_win:
        confirm_win.destroy()
    confirm_win = Gtk.Window()
    confirm_win.set_property("name", "popup")

    GtkLayerShell.init_for_window(confirm_win)
    GtkLayerShell.set_layer(confirm_win, GtkLayerShell.Layer.OVERLAY)
    # GtkLayerShell.set_keyboard_mode(confirm_win, GtkLayerShell.KeyboardMode.ON_DEMAND)

    confirm_win.set_resizable(False)
    confirm_win.set_modal(True)
    grid = Gtk.Grid()
    grid.set_column_spacing(12)
    grid.set_row_spacing(12)
    grid.set_column_homogeneous(True)
    grid.set_property("margin", 12)
    confirm_win.add(grid)
    lbl = Gtk.Label.new("{}?".format(voc["keep-current-settings"]))
    grid.attach(lbl, 0, 0, 2, 1)

    global counter
    counter = config["confirm-timeout"]

    cnt_lbl = Gtk.Label.new(str(counter))
    grid.attach(cnt_lbl, 0, 1, 2, 1)
    btn_restore = Gtk.Button.new_with_label(voc["restore"])

    btn_restore.connect("clicked", restore_old_settings, backup, path)

    grid.attach(btn_restore, 0, 2, 1, 1)
    btn_keep = Gtk.Button.new_with_label(voc["keep"])
    btn_keep.connect("clicked", keep_current_settings, config_dir, profile_name)
    grid.attach(btn_keep, 1, 2, 1, 1)

    confirm_win.show_all()

    global src_tag
    src_tag = GLib.timeout_add_seconds(1, count_down, cnt_lbl, backup, path)


def count_down(label, backup, path):
    global counter
    if counter > 0:
        counter -= 1
        label.set_text(str(counter))
        return True

    restore_old_settings(None, backup, path)


def keep_current_settings(btn, config_dir=None, profile_name=None):
    if src_tag > 0:
        GLib.Source.remove(src_tag)
    confirm_win.close()

    if os.getenv("NIRI_SOCKET"):
        niri_reload_config()
        GLib.timeout_add(2000, create_display_buttons)

    if config_dir and profile_name:
        if config.get("profile-bound-wallpapers", True):
            threading.Thread(
                target=WallpaperManager.apply_profile_wallpapers,
                args=(config_dir, profile_name),
                daemon=True,
            ).start()


def restore_old_settings(btn, backup, path):
    print("[Settings] Restoring old settings...")
    if src_tag > 0:
        GLib.Source.remove(src_tag)

    if os.getenv("SWAYSOCK"):
        save_list_to_text_file(backup, path)

        # Parse backup file back to commands and execute them
        single_line = ""
        # omit comments & empty lines
        for line in backup:
            if not line.startswith("#") and line:
                single_line += line
        # remove "{"
        single_line = single_line.replace("{", "")
        # convert multiple spaces into single
        single_line = " ".join(single_line.split())
        cmds = single_line.split("}")
        # execute line by line
        i3 = Connection()
        for cmd in cmds:
            if cmd:
                i3.command(cmd)

        confirm_win.close()
        create_display_buttons()

    elif os.getenv("HYPRLAND_INSTANCE_SIGNATURE"):
        save_list_to_text_file(backup, path)
        confirm_win.close()
        # Don't execute any command here, just save the file and wait for Hyprland to notice and apply the change.
        # Let's give it some time to do it before refreshing UI.
        GLib.timeout_add(2000, create_display_buttons)

    elif os.getenv("NIRI_SOCKET"):
        # For niri: backup is a file path (.bak), restore by copying it back
        if backup and os.path.isfile(backup):
            shutil.copy2(backup, path)
            print(f"[niri] Restored from {backup}")
        confirm_win.close()
        niri_reload_config()
        GLib.timeout_add(2000, create_display_buttons)


def main():
    GLib.set_prgname("nwg-displays")

    parser = argparse.ArgumentParser()

    if sway:
        parser.add_argument(
            "-o",
            "--outputs_path",
            type=str,
            default="{}/outputs".format(sway_config_dir),
            help="path to save Outputs config to, default: {}".format(
                "{}/outputs".format(sway_config_dir)
            ),
        )

        parser.add_argument(
            "-n",
            "--num_ws",
            type=int,
            default=8,
            help="number of Workspaces in use, default: 8",
        )

    elif hypr:
        parser.add_argument(
            "-m",
            "--monitors_path",
            type=str,
            default="{}/monitors.conf".format(hypr_config_dir),
            help="path to save the monitors.conf file to, default: {}".format(
                "{}/monitors.conf".format(hypr_config_dir)
            ),
        )

        parser.add_argument(
            "-w",
            "--workspaces_path",
            type=str,
            default="{}/workspaces.conf".format(hypr_config_dir),
            help="path to save the workspaces.conf file to, default: {}".format(
                "{}/workspaces.conf".format(hypr_config_dir)
            ),
        )

        parser.add_argument(
            "-n",
            "--num_ws",
            type=int,
            default=10,
            help="number of Workspaces in use, default: 10",
        )

    elif niri:
        parser.add_argument(
            "-m",
            "--monitors_path",
            type=str,
            default="{}/monitor.kdl".format(niri_config_dir),
            help="path to save the monitor.kdl file to, default: {}".format(
                "{}/monitor.kdl".format(niri_config_dir)
            ),
        )

        parser.add_argument(
            "-n",
            "--num_ws",
            type=int,
            default=10,
            help="number of Workspaces in use, default: 10",
        )

    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version="%(prog)s version {}".format(__version__),
        help="display version information",
    )
    args = parser.parse_args()

    load_vocabulary()

    global outputs_path
    global workspaces_path
    if sway:
        if os.path.isdir(sway_config_dir):
            outputs_path = args.outputs_path
        else:
            eprint("sway config directory not found!")
            outputs_path = ""
    elif hypr:
        if os.path.isdir(hypr_config_dir):
            outputs_path = args.monitors_path
            # 97, 115
            if os.path.lexists(outputs_path):
                is_writable = os.access(outputs_path, os.W_OK)
                if os.path.islink(outputs_path) and not is_writable:
                    eprint(f"INFO: '{outputs_path}' is a read-only symlink. Replacing with a writable file.")
                    tmp_path = f"{outputs_path}.tmp"
                    try:
                        with open(outputs_path, 'r') as src_file, open(tmp_path, 'w') as tmp_file:
                            tmp_file.write(src_file.read())
                        backup_path = f"{outputs_path}.bkp"
                        counter = 1
                        while os.path.lexists(backup_path):
                            backup_path = f"{outputs_path}.bkp{counter}"
                            counter += 1
                        eprint(f"INFO: Backing up '{outputs_path}' to '{backup_path}'")
                        os.rename(outputs_path, backup_path)
                        os.rename(tmp_path, outputs_path)
                    except Exception as e:
                        eprint(f"ERROR: Failed to replace read-only symlink: {e}")
                elif not os.path.islink(outputs_path) and not is_writable:
                    eprint(f"INFO: '{outputs_path}' is a read-only file. Making it writable.")
                    try:
                        os.chmod(outputs_path, os.stat(outputs_path).st_mode | stat.S_IWUSR)
                    except Exception as e:
                        eprint(f"ERROR: Failed to make file writable: {e}")

            workspaces_path = args.workspaces_path
        else:
            eprint("Hyprland config directory not found!")
            outputs_path = ""
            workspaces_path = ""
    elif niri:
        if os.path.isdir(niri_config_dir):
            outputs_path = args.monitors_path
            # Check if file is writable
            if os.path.lexists(outputs_path):
                is_writable = os.access(outputs_path, os.W_OK)
                if os.path.islink(outputs_path) and not is_writable:
                    eprint(f"INFO: '{outputs_path}' is a read-only symlink. Replacing with a writable file.")
                    tmp_path = f"{outputs_path}.tmp"
                    try:
                        with open(outputs_path, 'r') as src_file, open(tmp_path, 'w') as tmp_file:
                            tmp_file.write(src_file.read())
                        backup_path = f"{outputs_path}.bkp"
                        counter = 1
                        while os.path.lexists(backup_path):
                            backup_path = f"{outputs_path}.bkp{counter}"
                            counter += 1
                        eprint(f"INFO: Backing up '{outputs_path}' to '{backup_path}'")
                        os.rename(outputs_path, backup_path)
                        os.rename(tmp_path, outputs_path)
                    except Exception as e:
                        eprint(f"ERROR: Failed to replace read-only symlink: {e}")
                elif not os.path.islink(outputs_path) and not is_writable:
                    eprint(f"INFO: '{outputs_path}' is a read-only file. Making it writable.")
                    try:
                        os.chmod(outputs_path, os.stat(outputs_path).st_mode | stat.S_IWUSR)
                    except Exception as e:
                        eprint(f"ERROR: Failed to make file writable: {e}")
        else:
            eprint("niri config directory not found!")
            outputs_path = ""

    global num_ws
    num_ws = args.num_ws
    if sway:
        print("[Info] Number of workspaces: {}".format(num_ws))

    config_file = os.path.join(config_dir, "config")
    global config
    if not os.path.isfile(config_file):
        # migrate old config file, if not yet migrated
        if os.path.isfile(os.path.join(old_config_dir, "config")):
            print("[Config] Migrating config to the proper path...")
            os.rename(old_config_dir, config_dir)
        else:
            if not os.path.isdir(config_dir):
                os.makedirs(config_dir, exist_ok=True)
            print("[Config] '{}' file not found, creating default".format(config_file))
            save_json(config, config_file)
    else:
        config = load_json(config_file)

    if config_keys_missing(config, config_file):
        config = load_json(config_file)

    if "profile-bound-wallpapers" not in config:
        config["profile-bound-wallpapers"] = True
        save_json(config, config_file)

    # Initialize the profile manager
    global profile_manager
    profile_manager = ProfileManager(config_dir, config, voc)

    global snap_threshold_scaled
    snap_threshold_scaled = config["snap-threshold"]

    builder = Gtk.Builder()
    builder.add_from_file(os.path.join(dir_name, "resources/main.glade"))

    window = builder.get_object("window")
    screen = Gdk.Screen.get_default()
    provider = Gtk.CssProvider()
    style_context = Gtk.StyleContext()
    style_context.add_provider_for_screen(
        screen, provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
    )
    try:
        file = os.path.join(dir_name, "resources/style.css")
        provider.load_from_path(file)
    except:
        sys.stderr.write(
            "ERROR: {} file not found, using GTK styling\n".format(
                os.path.join(dir_name, "resources/style.css")
            )
        )

    window.connect("key-release-event", handle_keyboard)
    window.connect("destroy", Gtk.main_quit)

    builder.get_object("lbl-modes").set_label("{}:".format(voc["modes"]))
    builder.get_object("lbl-position-x").set_label("{}:".format(voc["position-x"]))
    builder.get_object("lbl-refresh").set_label("{}:".format(voc["refresh"]))
    builder.get_object("lbl-scale").set_label("{}:".format(voc["scale"]))
    builder.get_object("lbl-scale-filter").set_label("{}:".format(voc["scale-filter"]))
    builder.get_object("lbl-size").set_label("{}:".format(voc["size"]))
    builder.get_object("lbl-transform").set_label("{}:".format(voc["transform"]))
    builder.get_object("lbl-zoom").set_label("{}:".format(voc["zoom"]))

    global form_name
    form_name = builder.get_object("name")

    global form_description
    form_description = builder.get_object("description")

    global form_dpms
    form_dpms = builder.get_object("dpms")
    form_dpms.set_tooltip_text(voc["dpms-tooltip"])
    form_dpms.connect("toggled", on_dpms_toggled)
    # if sway:
    #     form_dpms.set_tooltip_text(voc["dpms-tooltip"])
    #     form_dpms.connect("toggled", on_dpms_toggled)
    # else:
    #     form_dpms.set_sensitive(False)

    global form_adaptive_sync
    form_adaptive_sync = builder.get_object("adaptive-sync")
    if sway:
        form_adaptive_sync.set_label(voc["adaptive-sync"])
        form_adaptive_sync.set_tooltip_text(voc["adaptive-sync-tooltip"])
        form_adaptive_sync.connect("toggled", on_adaptive_sync_toggled)
    else:
        form_adaptive_sync.set_sensitive(False)

    global form_custom_mode
    form_custom_mode = builder.get_object("custom-mode")
    if sway:
        form_custom_mode.set_label(voc["custom-mode"])
        form_custom_mode.set_tooltip_text(voc["custom-mode-tooltip"])
        form_custom_mode.connect("toggled", on_custom_mode_toggle)
    else:
        form_custom_mode.set_sensitive(False)

    global form_view_scale
    form_view_scale = builder.get_object("view-scale")
    form_view_scale.set_tooltip_text(voc["view-scale-tooltip"])
    adj = Gtk.Adjustment(
        lower=0.1, upper=0.6, step_increment=0.05, page_increment=0.1, page_size=0.1
    )
    form_view_scale.configure(adj, 1, 2)
    form_view_scale.connect("changed", on_view_scale_changed)

    global form_x
    form_x = builder.get_object("x")
    adj = Gtk.Adjustment(
        lower=0, upper=60000, step_increment=1, page_increment=10, page_size=1
    )
    form_x.configure(adj, 1, 0)
    form_x.connect("value-changed", on_pos_x_changed)

    global form_y
    form_y = builder.get_object("y")
    adj = Gtk.Adjustment(
        lower=0, upper=40000, step_increment=1, page_increment=10, page_size=1
    )
    form_y.configure(adj, 1, 0)
    form_y.connect("value-changed", on_pos_y_changed)

    global form_width
    form_width = builder.get_object("width")
    adj = Gtk.Adjustment(
        lower=0, upper=7680, step_increment=1, page_increment=10, page_size=1
    )
    form_width.configure(adj, 1, 0)
    form_width.connect("value-changed", on_width_changed)

    global form_height
    form_height = builder.get_object("height")
    adj = Gtk.Adjustment(
        lower=0, upper=4320, step_increment=1, page_increment=10, page_size=1
    )
    form_height.configure(adj, 1, 0)
    form_height.connect("value-changed", on_height_changed)

    global form_scale
    form_scale = builder.get_object("scale")
    adj = Gtk.Adjustment(
        lower=0.1, upper=10, step_increment=0.1, page_increment=10, page_size=1
    )
    form_scale.configure(adj, 0.1, 6)
    form_scale.connect("value-changed", on_scale_changed)

    global form_scale_filter
    form_scale_filter = builder.get_object("scale-filter")
    if sway:
        form_scale_filter.set_tooltip_text(voc["scale-filter-tooltip"])
        form_scale_filter.connect("changed", on_scale_filter_changed)
    else:
        form_scale_filter.set_sensitive(False)

    global form_refresh
    form_refresh = builder.get_object("refresh")
    adj = Gtk.Adjustment(
        lower=1, upper=1200, step_increment=1, page_increment=10, page_size=1
    )
    form_refresh.configure(adj, 1, 3)
    form_refresh.connect("changed", on_refresh_changed)

    global form_modes
    form_modes = builder.get_object("modes")
    form_modes.set_wrap_width(1)
    form_modes.set_tooltip_text(voc["modes-tooltip"])
    form_modes.connect("changed", on_mode_changed)

    global form_use_desc
    form_use_desc = builder.get_object("use-desc")
    form_use_desc.set_label("{}".format(voc["use-desc"]))
    form_use_desc.set_tooltip_text("{}".format(voc["use-desc-tooltip"]))
    form_use_desc.connect("toggled", on_use_desc_toggled)

    global form_transform
    form_transform = builder.get_object("transform")
    form_transform.set_wrap_width(1)
    form_transform.set_tooltip_text(voc["transform-tooltip"])
    form_transform.connect("changed", on_transform_changed)

    global form_wrapper_box
    form_wrapper_box = builder.get_object("wrapper-box")

    global form_workspaces
    form_workspaces = builder.get_object("workspaces")
    form_workspaces.set_label(voc["workspaces"])
    form_workspaces.set_tooltip_text(voc["workspaces-tooltip"])
    if sway:
        form_workspaces.connect("clicked", create_workspaces_window)
    elif hypr:
        form_workspaces.connect("clicked", create_workspaces_window_hypr)
    elif niri:
        # Niri uses dynamic workspaces, no need for workspace assignment
        form_workspaces.set_sensitive(False)
        form_workspaces.set_tooltip_text("Niri uses dynamic workspaces")

    global form_close
    form_close = builder.get_object("close")
    form_close.set_label(voc["close"])
    form_close.connect("clicked", Gtk.main_quit)
    form_close.grab_focus()

    global form_apply
    form_apply = builder.get_object("apply")
    form_apply.set_label(voc["apply"])
    if (sway and sway_config_dir) or (hypr and hypr_config_dir) or (niri and niri_config_dir):
        form_apply.connect("clicked", on_apply_button, profile_manager)
    else:
        form_apply.set_sensitive(False)
        form_apply.set_tooltip_text("Config dir not found")

    global form_version
    form_version = builder.get_object("version")
    form_version.set_text("v{}".format(__version__))

    wrapper = builder.get_object("wrapper")
    wrapper.set_property("name", "wrapper")

    global fixed
    fixed = builder.get_object("fixed")

    create_display_buttons()

    global outputs_activity
    outputs_activity = list_outputs_activity()
    lbl = Gtk.Label()
    lbl.set_text("{}:".format(voc["active"]))
    form_wrapper_box.pack_start(lbl, False, False, 3)
    for key in outputs_activity:
        cb = Gtk.CheckButton()
        cb.set_label(key)
        cb.set_active(outputs_activity[key])
        cb.connect("toggled", on_output_toggled, key)
        form_wrapper_box.pack_start(cb, False, False, 3)

    btn = Gtk.Button.new_with_label(voc["toggle"])
    if sway:
        btn.set_tooltip_text(voc["toggle-tooltip"])
        btn.connect("clicked", on_toggle_button)
        form_wrapper_box.pack_start(btn, False, False, 3)
    else:
        btn.destroy()

    if hypr or niri:
        grid = builder.get_object("grid")

        global form_ten_bit
        form_ten_bit = Gtk.CheckButton.new_with_label(voc["10-bit-support"])
        form_ten_bit.set_tooltip_text(voc["10-bit-support-tooltip"])
        form_ten_bit.connect("toggled", on_ten_bit_toggled)
        grid.attach(form_ten_bit, 5, 4, 1, 1)

        lbl = Gtk.Label.new("Mirror:")
        lbl.set_property("halign", Gtk.Align.END)
        grid.attach(lbl, 6, 4, 1, 1)

        global form_mirror
        form_mirror = Gtk.ComboBoxText()
        form_mirror.set_wrap_width(1)
        form_mirror.connect("changed", on_mirror_selected)
        grid.attach(form_mirror, 7, 4, 1, 1)

    # Add profile management buttons
    separator = Gtk.Separator(orientation=Gtk.Orientation.VERTICAL)
    form_wrapper_box.pack_start(separator, False, False, 10)

    profile_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=6)
    form_wrapper_box.pack_start(profile_box, False, False, 0)

    # Add label for current profile name
    profile_label = Gtk.Label()
    profile_label.set_text(
        voc.get("current-profile", "Profile") + ": " + voc.get("none", "None")
    )
    profile_label.set_property("margin-end", 10)
    profile_label.set_property("margin-start", 5)
    profile_box.pack_start(profile_label, False, False, 0)

    btn_new_profile = Gtk.Button.new_with_label(voc.get("new", "New"))
    btn_new_profile.set_tooltip_text(
        voc.get("new-profile-tooltip", "Create a new profile")
    )
    btn_new_profile.connect("clicked", profile_manager.create_profile)
    profile_box.pack_start(btn_new_profile, False, False, 0)

    btn_select_profile = Gtk.Button.new_with_label(voc.get("select", "Select"))
    btn_select_profile.set_tooltip_text(
        voc.get("select-profile-tooltip", "Select a profile")
    )
    btn_select_profile.connect("clicked", profile_manager.select_profile)
    profile_box.pack_start(btn_select_profile, False, False, 0)

    btn_save_profile = Gtk.Button.new_with_label(voc.get("save", "Save"))
    btn_save_profile.set_tooltip_text(
        voc.get("save-profile-tooltip", "Save to current profile")
    )
    btn_save_profile.connect("clicked", profile_manager.save_profile)
    btn_save_profile.set_sensitive(
        False
    )  # Initially disabled until a profile is selected
    profile_box.pack_start(btn_save_profile, False, False, 0)

    global form_profile_wallpapers
    form_profile_wallpapers = Gtk.CheckButton.new_with_label(
        voc.get("profile-bound-wallpapers", "Profile-bound wallpapers")
    )
    form_profile_wallpapers.set_tooltip_text(
        voc.get(
            "profile-bound-wallpapers-tooltip",
            "Save and load wallpapers together with display profiles",
        )
    )
    form_profile_wallpapers.set_active(config.get("profile-bound-wallpapers", True))
    form_profile_wallpapers.connect("toggled", on_profile_wallpapers_toggled)
    form_wrapper_box.pack_start(form_profile_wallpapers, False, False, 6)

    # Register the save button with the profile manager
    profile_manager.set_save_button(btn_save_profile)

    # Register the profile label with the profile manager
    profile_manager.set_profile_label(profile_label)

    # Also pass display_buttons and other required data to profile manager
    profile_manager.set_display_buttons(display_buttons)
    profile_manager.set_fixed(fixed)
    profile_manager.set_update_callback(update_form_from_widget)

    if display_buttons:
        update_form_from_widget(display_buttons[0])
        display_buttons[0].select()

    screen = Gdk.Screen.get_default()
    provider = Gtk.CssProvider()
    style_context = Gtk.StyleContext()
    style_context.add_provider_for_screen(
        screen, provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
    )
    css = b""" #popup { border-radius: 6px; border: solid 1px; border-color: #f00 } """
    provider.load_from_data(css)

    window.show_all()

    # Gtk.Fixed does not respect expand properties. That's why we need
    # to scale the window automagically if opened as a floating_con
    Gdk.threads_add_timeout(GLib.PRIORITY_LOW, 100, scale_if_floating)

    Gtk.main()


if __name__ == "__main__":
    sys.exit(main())



import os
import stat
import subprocess
import json
import time

# from nwg_displays.main import sway
from nwg_displays.tools import is_command, load_text_file


class WallpaperManager:
    @staticmethod
    def get_current_wallpapers():
        """Returns a dict containing path and mode fields attached to monitor names. For swww mode fields are empty."""
        if not is_command("nwg-shell") and not is_command("swww"):
            return {}

        # Firstly we need to support nwg-shell-related wallpapers, and we use Azote Wallpaper Manager here.
        # Let's parse the ~/.azotebg or ~/azotebg-hyprland batch file.
        if is_command("nwg-shell") and (os.getenv("SWAYSOCK") or os.getenv("HYPRLAND_INSTANCE_SIGNATURE")):
            if os.getenv("SWAYSOCK"):
                azotebg_file = os.path.join(os.getenv("HOME"), ".azotebg")
            else:
                azotebg_file = os.path.join(os.getenv("HOME"), ".azotebg-hyprland")
            azotebg_content = load_text_file(azotebg_file)
            if azotebg_content:
                return WallpaperManager.parse_azotebg_content(azotebg_content)
            else:
                print(f"[Error] Couldn't find ~/.azotebg* file")
                return {}
        else:
            try:
                output = subprocess.check_output(["swww", "query"], text=True)
                return WallpaperManager._parse_swww_output(output)
            except Exception as e:
                print(f"[Error] Failed to query swww: {e}")
                return {}

    @staticmethod
    def parse_azotebg_content(content):
        """Parses the azotebg content and returns a dict containing path and mode for each output."""
        # The content of the batch file written by Azote will look something like this:
        # pkill swaybg
        # swaybg -o 'DP-1' -i "/home/piotr/Obrazy/Wallpapers/nwg_3a_3840x2160px.jpg" -m fill &
        # swaybg -o 'DP-2' -i "/home/piotr/Obrazy/Wallpapers/nwg_3a_3840x2160px.jpg" -m fill &
        current_walls = {}
        for line in content.splitlines():
            fields = line.split()
            if line.startswith("swaybg"):
                try:
                    monitor = fields[2][1:-1]
                    path = fields[4][1:-1]
                    mode = fields[6]
                    current_walls[monitor] = {"path": path, "mode": mode}
                except IndexError:
                    print(f"[Error parsing line] {line}")
        return current_walls


    @staticmethod
    def _parse_swww_output(output):
        current_walls = {}
        for line in output.splitlines():
            if "currently displaying: " in line:
                # New swww format: <namespace>: <monitor>: <resolution>, ... currently displaying: <path>
                path = line.split("currently displaying: ")[-1].strip()
                if path.startswith("image: "):
                    path = path[7:]
                parts = line.split(": ")
                monitor = None
                # Look for the resolution field (starts with digit) and take the preceding field as monitor
                for i, part in enumerate(parts):
                    if len(part) > 0 and part[0].isdigit() and i > 0:
                        candidate = parts[i - 1].strip()
                        if candidate != "scale":
                            monitor = candidate
                            break
                if monitor:
                    # for the sake of Azote compatibility, we also need to store the mode value, that will be empty here
                    current_walls[monitor] = {"path": path, "mode": ""}
            elif ": " in line:
                parts = line.split(": ", 1)
                if len(parts) == 2:
                    monitor = parts[0].strip()
                    if monitor.startswith("Output "):
                        monitor = monitor[7:]
                    # swww sometimes returns extra info, ensure we just get the path
                    path = parts[1].strip().split(" ")[0]
                    current_walls[monitor] = {"path": path, "mode": ""}
        return current_walls

    @staticmethod
    def apply_wallpapers(wallpaper_data):
        if not wallpaper_data:
            return

        if is_command("nwg-shell"):
            WallpaperManager._apply_azotebg(wallpaper_data)
            return

        if is_command("swww"):
            WallpaperManager._apply_swww(wallpaper_data)
            return

        if is_command("hyprpaper"):
            WallpaperManager._apply_hyprpaper()
            return

        print("[Error] No wallpaper daemon found (swww/hyprpaper/swaybg)")

    @staticmethod
    def _apply_azotebg(wallpaper_data):
        print("[Wallpapers] Using nwg-shell/Azote/swaybg backend")
        batch_content = ['#!/usr/bin/env bash', 'pkill swaybg']
        for key in wallpaper_data:
            batch_content.append(f"swaybg -o '{key}' -i \"{wallpaper_data[key]['path']}\" -m {wallpaper_data[key]['mode']} &")
        print("\n".join(batch_content))

        if os.getenv("SWAYSOCK"):
            azotebg_file = os.path.join(os.getenv("HOME"), ".azotebg")
        else:
            azotebg_file = os.path.join(os.getenv("HOME"), ".azotebg-hyprland")

        # write to .azotebg* file
        with open(azotebg_file, 'w') as f:
            for item in batch_content:
                f.write("%s\n" % item)
        # make the file executable
        st = os.stat(azotebg_file)
        os.chmod(azotebg_file, st.st_mode | stat.S_IEXEC)
        # execute
        subprocess.call(azotebg_file, shell=True)

    @staticmethod
    def _apply_swww(wallpaper_data):
        print("[Wallpapers] Using swww backend")
        # for monitor, path in wallpaper_data.items():
        for key in wallpaper_data:
            path = wallpaper_data[key]["path"]
            path = os.path.expanduser(os.path.expandvars(path))
            if os.path.isfile(path):
                print(f"[Wallpapers] Setting {key} to {path}")
                # Using 'grow' transition for smooth profile switching
                cmd = f"swww img -o {key} '{path}' --transition-type grow --transition-pos 0.8,0.9 --transition-step 90"
                os.system(cmd)
            else:
                print(f"[Warning] Wallpaper file not found: {path}")

    @staticmethod
    def _apply_hyprpaper():
        print("[Wallpapers] Hyprpaper detected (swww is recommended for profiles)")
        pass

    @staticmethod
    def apply_profile_wallpapers(config_dir, profile_name):
        profile_path = os.path.join(config_dir, "profiles", f"{profile_name}.json")
        if os.path.isfile(profile_path):
            try:
                with open(profile_path, "r") as f:
                    profile_data = json.load(f)
                if "wallpapers" in profile_data:
                    print("[Profile] Applying wallpapers...")
                    time.sleep(1)
                    WallpaperManager.apply_wallpapers(profile_data["wallpapers"])
            except Exception as e:
                print(f"[Error] Failed to apply wallpapers from profile: {e}")



# !/usr/bin/env python3
import datetime
import json
import os
import shutil
import socket
import subprocess
import sys

import gi

gi.require_version('Gdk', '3.0')
from gi.repository import Gdk

if os.getenv("SWAYSOCK"):
    from i3ipc import Connection


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


def get_config_home():
    xdg_config_home = os.getenv('XDG_CONFIG_HOME')
    config_home = xdg_config_home if xdg_config_home else os.path.join(
        os.getenv("HOME"), ".config")

    return config_home


def get_config_dir():
    xdg_config_home = os.getenv("XDG_CONFIG_HOME", os.path.expanduser("~/.config"))
    return os.path.join(xdg_config_home, "nwg-displays")


def niri_msg(cmd):
    """Execute niri msg command and return output"""
    niri_socket = os.getenv("NIRI_SOCKET")
    if not niri_socket:
        return None
    
    if not os.path.exists(niri_socket):
        eprint(f"[niri] Socket file not found: {niri_socket}")
        return None
    
    try:
        s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        s.connect(niri_socket)
        request = cmd + "\n"
        s.send(request.encode("utf-8"))
        s.shutdown(socket.SHUT_WR)
        output = ""
        while True:
            chunk = s.recv(4096)
            if not chunk:
                break
            output += chunk.decode('utf-8')
        s.close()
        return output.strip()
    except Exception as e:
        eprint(f"Error communicating with niri: {e}")
        return None


def niri_reload_config():
    """Reload niri configuration via niri msg CLI"""
    try:
        subprocess.run(['niri', 'msg', 'action', 'load-config-file', 
                       os.path.join(get_config_home(), "niri", "config.kdl")],
                      capture_output=True, text=True, timeout=5)
        print("[niri] Config reloaded")
    except Exception as e:
        eprint(f"[niri] Failed to reload config: {e}")


def get_config():
    config_dir = get_config_dir()
    config_file = os.path.join(config_dir, "config")

    if not os.path.isfile(config_file):
        print(f"[Error] Config file not found at {config_file}")
        sys.exit(1)

    config = load_json(config_file)
    if config is None:
        print("[Error] Failed to load configuration")
        sys.exit(1)

    return config, config_file


def hyprctl(cmd):
    # /tmp/hypr moved to $XDG_RUNTIME_DIR/hypr in #5788
    xdg_runtime_dir = os.getenv("XDG_RUNTIME_DIR")
    hypr_dir = f"{xdg_runtime_dir}/hypr" if xdg_runtime_dir and os.path.isdir(
        f"{xdg_runtime_dir}/hypr") else "/tmp/hypr"

    s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    s.connect(f"{hypr_dir}/{os.getenv('HYPRLAND_INSTANCE_SIGNATURE')}/.socket.sock")

    s.send(cmd.encode("utf-8"))
    output = s.recv(20480).decode('utf-8')
    s.close()

    return output


def is_command(cmd):
    return shutil.which(cmd) is not None


def list_outputs():
    if os.getenv("NIRI_SOCKET"):
        outputs_dict = {}
        eprint("Running on niri")
        
        try:
            try:
                result = subprocess.run(['niri', 'msg', '-j', 'outputs'], 
                                      capture_output=True, text=True, timeout=5)
                output = result.stdout
            except Exception as e:
                eprint(f"[niri] Failed to run niri msg: {e}")
                # Fallback to socket communication
                output = niri_msg('{"Outputs":null}')
                
            if not output:
                eprint("[niri] Could not communicate with niri compositor")
                return {}
                
            data = json.loads(output)
            
            # niri msg returns dict directly: {"eDP-1": {...}}
            # Or via IPC: {"Ok":{"Outputs":{"eDP-1":{...}}}}
            monitors_dict = None
            if isinstance(data, dict):
                if "Ok" in data:
                    monitors_dict = data.get("Outputs", {})
                elif "Err" in data:
                    eprint(f"[niri] Error from compositor: {data['Err']}")
                    return {}
                else:
                    # Direct response from niri msg
                    monitors_dict = data
            
            if monitors_dict is None or not isinstance(monitors_dict, dict):
                eprint(f"[niri] Unexpected response format: {data}")
                return {}
            
            # monitors_dict is a dict keyed by output name
            transforms = {"Normal": "normal", "90": "90", "180": "180", "270": "270", 
                         "Flipped": "flipped", "Flipped90": "flipped-90", 
                         "Flipped180": "flipped-180", "Flipped270": "flipped-270"}
            
            for name, mon in monitors_dict.items():
                # Get current mode info
                current_mode_idx = mon.get("current_mode", 0)
                modes_list = mon.get("modes", [])
                current_mode = modes_list[current_mode_idx] if modes_list and current_mode_idx < len(modes_list) else {}
                
                logical = mon.get("logical", {})
                
                # Store raw make/model for accurate matching (not just description)
                raw_make = mon.get("make", "")
                raw_model = mon.get("model", "")
                raw_serial = mon.get("serial")
                physical_size = mon.get("physical_size", [])
                
                # Format description for backward compatibility
                description = f'{raw_make} {raw_model} {raw_serial or ""}'.strip()
                
                outputs_dict[name] = {
                    "active": True,  # If it's in the list, it's active
                    "dpms": True,
                    "description": description,
                    "x": int(logical.get("x", 0)),
                    "y": int(logical.get("y", 0)),
                    "logical-width": int(logical.get("width", 0)),
                    "logical-height": int(logical.get("height", 0)),
                    "physical-width": int(current_mode.get("width", 0)),
                    "physical-height": int(current_mode.get("height", 0)),
                    "transform": transforms.get(logical.get("transform", "Normal"), "normal"),
                    "scale": float(logical.get("scale", 1.0)),
                    "scale_filter": "linear",
                    "refresh": round(float(current_mode.get("refresh_rate", 60000)) / 1000, 2),
                    "modes": [],
                    "focused": mon.get("is_focused", False),
                    "adaptive_sync_status": "enabled" if mon.get("vrr_enabled", False) else "disabled",
                    "mirror": "",
                    "ten_bit": False,
                    "monitor": None,
                    # Store raw make/model for matching
                    "__niri_make": raw_make,
                    "__niri_model": raw_model,
                    "__niri_serial": raw_serial,
                    "__niri_physical_size": physical_size,
                }
                
                # Parse available modes
                for mode in modes_list:
                    try:
                        outputs_dict[name]["modes"].append({
                            "width": int(mode.get("width", 0)),
                            "height": int(mode.get("height", 0)),
                            "refresh": float(mode.get("refresh_rate", 60000))
                        })
                    except:
                        pass
        except json.JSONDecodeError as e:
            eprint(f"[niri] JSON decode error: {e}")
            return {}
        except Exception as e:
            eprint(f"Error parsing niri outputs: {e}")
            # Return empty dict if parsing fails
            return {}
            
    elif os.getenv("SWAYSOCK"):
        outputs_dict = {}
        eprint("Running on sway")
        i3 = Connection()
        tree = i3.get_tree()
        for item in tree:
            if item.type == "output" and not item.name.startswith("__"):
                outputs_dict[item.name] = {"x": item.rect.x,
                                           "y": item.rect.y,
                                           "logical-width": item.rect.width,
                                           "logical-height": item.rect.height,
                                           "physical-width": item.ipc_data["current_mode"]["width"],
                                           "physical-height": item.ipc_data["current_mode"]["height"]}

                outputs_dict[item.name]["active"] = item.ipc_data["active"]
                outputs_dict[item.name]["dpms"] = item.ipc_data["dpms"]
                outputs_dict[item.name]["transform"] = item.ipc_data[
                    "transform"] if "transform" in item.ipc_data else None
                outputs_dict[item.name]["scale"] = float(item.ipc_data["scale"]) if "scale" in item.ipc_data else None
                outputs_dict[item.name]["scale_filter"] = item.ipc_data["scale_filter"]
                outputs_dict[item.name]["adaptive_sync_status"] = item.ipc_data["adaptive_sync_status"]
                outputs_dict[item.name]["refresh"] = \
                    item.ipc_data["current_mode"]["refresh"] / 1000 if "refresh" in item.ipc_data[
                        "current_mode"] else None
                outputs_dict[item.name]["modes"] = item.ipc_data["modes"] if "modes" in item.ipc_data else []
                outputs_dict[item.name]["description"] = "{} {} {}".format(item.ipc_data["make"],
                                                                           item.ipc_data["model"],
                                                                           item.ipc_data["serial"])
                outputs_dict[item.name]["focused"] = item.ipc_data["focused"]

                outputs_dict[item.name]["mirror"] = ""  # We only use it on Hyprland
                outputs_dict[item.name]["ten_bit"] = False  # We have no way to check it on sway
                outputs_dict[item.name]["monitor"] = None

    elif os.getenv("HYPRLAND_INSTANCE_SIGNATURE"):
        monitors_all = json.loads(hyprctl("j/monitors all"))
        monitors = json.loads(hyprctl("j/monitors"))
        active = []
        for item in monitors:
            active.append(item["name"])
        outputs_dict = {}
        for mon in monitors_all:
            name = mon["name"]
            outputs_dict[name] = {"active": True} if name in active else {"active": False}

        eprint("Running on Hyprland")

        # 1. Mirroring is impossible to check in any way. We need to parse back the monitors.conf file, and it sucks.
        mirrors = {}
        hypr_config_dir = os.path.join(get_config_home(), "hypr")
        monitors_file = os.path.join(hypr_config_dir, "monitors.conf")
        if os.path.isfile(monitors_file):
            lines = load_text_file(monitors_file).splitlines()
            for line in lines:
                if line and not line.startswith("#"):  # skip comments
                    if "mirror" in line:
                        settings = line.split("=")[1].split(",")
                        mirrors[settings[0].strip()] = settings[-1].strip()

        # 2. This won't work w/ Hyprland <= 0.36.0
        output = hyprctl("j/monitors all")
        monitors = json.loads(output)
        transforms = {0: "normal", 1: "90", 2: "180", 3: "270", 4: "flipped", 5: "flipped-90", 6: "flipped-180",
                      7: "flipped-270"}
        for m in monitors:
            outputs_dict[m["name"]]["mirror"] = mirrors[name] if name in mirrors else ""

            outputs_dict[m["name"]]["scale_filter"] = None
            outputs_dict[m["name"]]["modes"] = []
            outputs_dict[m["name"]]["focused"] = m["focused"]
            outputs_dict[m["name"]]["adaptive_sync_status"] = "enabled" if m["vrr"] else "disabled"

            outputs_dict[m["name"]]["description"] = f'{m["description"]}'
            outputs_dict[m["name"]]["x"] = int(m["x"])
            outputs_dict[m["name"]]["y"] = int(m["y"])

            outputs_dict[m["name"]]["refresh"] = round(m["refreshRate"], 2)

            outputs_dict[m["name"]]["logical-width"] = m["width"] / m["scale"]
            outputs_dict[m["name"]]["logical-height"] = m["height"] / m["scale"]

            outputs_dict[m["name"]]["physical-width"] = m["width"]
            outputs_dict[m["name"]]["physical-height"] = m["height"]

            outputs_dict[m["name"]]["transform"] = transforms[m["transform"]]
            outputs_dict[m["name"]]["scale"] = m["scale"]
            outputs_dict[m["name"]]["focused"] = m["focused"]
            outputs_dict[m["name"]]["dpms"] = m["dpmsStatus"]

            outputs_dict[name]["modes"] = []

            for item in m["availableModes"]:
                line = item[:-2]  # split "Hz"
                w_h, r = line.split("@")
                w, h = w_h.split("x")
                try:
                    mode = {"width": int(w), "height": int(h), "refresh": float(r) * 1000}
                except ValueError as e:
                    eprint(e)
                outputs_dict[m["name"]]["modes"].append(mode)

            outputs_dict[m["name"]]["ten_bit"] = True if m["currentFormat"] in ["XRGB2101010", "XBGR2101010"] else False

            # to identify Gdk.Monitor
            outputs_dict[m["name"]]["model"] = m["model"]

            outputs_dict[m["name"]]["monitor"] = None

    else:
        eprint("This program only supports sway, Hyprland and niri, and we seem to be elsewhere, terminating.")
        sys.exit(1)

    # We used to assign Gdk.Monitor to output on the basis of x and y coordinates, but it no longer works,
    # starting from gtk3-1:3.24.42: all monitors have x=0, y=0. This is most likely a bug, but from now on
    # we must rely on gdk monitors order.
    # For niri, we use property-based matching to avoid order issues after config reload.
    monitors = []
    display = Gdk.Display.get_default()
    for i in range(display.get_n_monitors()):
        monitor = display.get_monitor(i)
        monitors.append(monitor)

    if os.getenv("NIRI_SOCKET"):
        # Niri: Use position-based matching with Gdk.Monitor geometry
        # This ensures correct matching even when multiple monitors have identical properties
        for key in outputs_dict:
            if key in outputs_dict:
                data = outputs_dict[key]
                output_x = data.get("x", 0)
                output_y = data.get("y", 0)
                
                # Find matching monitor by position (x, y coordinates)
                matched = False
                for m in monitors:
                    try:
                        geom = m.get_geometry()
                        gdk_x = geom.x
                        gdk_y = geom.y
                        # Check if positions match (allow small tolerance)
                        if abs(gdk_x - output_x) < 10 and abs(gdk_y - output_y) < 10:
                            outputs_dict[key]["monitor"] = m
                            matched = True
                            break
                    except:
                        pass
                
                # Fallback to index-based if position matching fails
                if not matched:
                    idx = 0
                    for k in outputs_dict:
                        if k == key:
                            break
                        idx += 1
                    if idx < len(monitors):
                        outputs_dict[key]["monitor"] = monitors[idx]
    else:
        # Sway/Hyprland: Keep original index-based matching
        idx = 0
        for key in outputs_dict:
            try:
                outputs_dict[key]["monitor"] = monitors[idx]
            except IndexError:
                print(f"Couldn't assign a Gdk.Monitor to {outputs_dict[key]}")
            idx += 1

    for key in outputs_dict:
        eprint(key, outputs_dict[key])
    return outputs_dict


def list_outputs_activity():
    result = {}
    if os.getenv("NIRI_SOCKET"):
        outputs = list_outputs()
        for name in outputs:
            result[name] = outputs[name].get("active", True)
    elif os.getenv("SWAYSOCK"):
        i3 = Connection()
        outputs = i3.get_outputs()
        for o in outputs:
            result[o.name] = o.active

    elif os.getenv("HYPRLAND_INSTANCE_SIGNATURE"):
        monitors_all = json.loads(hyprctl("j/monitors all"))
        monitors = json.loads(hyprctl("j/monitors"))
        active = []
        for item in monitors:
            active.append(item["name"])

        for mon in monitors_all:
            name = mon["name"]
            result[name] = True if name in active else False

    return result


def max_window_height():
    if os.getenv("SWAYSOCK"):
        i3 = Connection()
        outputs = i3.get_outputs()
        for o in outputs:
            if o.focused:
                if o.rect.width > o.rect.height:
                    return o.rect.height * 0.9
                else:
                    return o.rect.height / 2 * 0.9
    return None


def scale_if_floating():
    pid = os.getpid()
    if os.getenv("SWAYSOCK"):
        i3 = Connection()
        node = i3.get_tree().find_by_pid(pid)[0]
        if node.type == "floating_con":
            h = int(max_window_height())
            if h:
                i3.command("resize set height {}".format(h))


def min_val(a, b):
    if b < a:
        return b
    return a


def max_val(a, b):
    if b > a:
        return b
    return a


def round_down_to_multiple(i, m):
    return i / m * m


def round_to_nearest_multiple(i, m):
    if i % m > m / 2:
        return (i / m + 1) * m
    return i / m * m


def orientation_changed(transform, transform_old):
    return (is_rotated(transform) and not is_rotated(transform_old)) or (
            is_rotated(transform_old) and not is_rotated(transform))


def is_rotated(transform):
    return "90" in transform or "270" in transform


def inactive_output_description(name):
    if os.getenv("SWAYSOCK"):
        i3 = Connection()
        for item in i3.get_outputs():
            if item.name == name:
                return "{} {} {}".format(item.ipc_data["make"], item.ipc_data["model"],
                                         item.ipc_data["serial"])
    return None


def config_keys_missing(config, config_file):
    key_missing = False
    defaults = {
        "view-scale": 0.15,
        "snap-threshold": 10,
        "indicator-timeout": 500,
        "custom-mode": [],
        "use-desc": False,
        "confirm-timeout": 10,
        "profile-bound-wallpapers": True,
    }
    for key in defaults:
        if key not in config:
            config[key] = defaults[key]
            print("Added missing config key: '{}'".format(key), file=sys.stderr)
            key_missing = True

    if key_missing:
        save_json(config, config_file)

    return key_missing


def load_json(path):
    try:
        with open(path, 'r') as f:
            return json.load(f)
    except Exception as e:
        print("Error loading json: {}".format(e))
        return None


def save_json(src_dict, path):
    with open(path, 'w') as f:
        json.dump(src_dict, f, indent=2)


def save_list_to_text_file(data, file_path):
    text_file = open(file_path, "w")
    for line in data:
        text_file.write(line + "\n")
    text_file.close()


def save_kdl_output(data, file_path):
    """Save output configuration in KDL format for niri"""
    text_file = open(file_path, "w")
    now = datetime.datetime.now()
    line = "// Generated by nwg-displays on {} at {}. Do not edit manually.\n".format(
        datetime.datetime.strftime(now, '%Y-%m-%d'),
        datetime.datetime.strftime(now, '%H:%M:%S'))
    text_file.write(line + "\n")
    
    for d in data:
        name = d["name"]
        if not d["active"]:
            text_file.write(f'output "{name}" {{\n    off\n}}\n\n')
            continue
        
        text_file.write(f'output "{name}" {{\n')
        
        # Mode
        text_file.write(f'    mode "{d["physical_width"]}x{d["physical_height"]}@{d["refresh"]}"\n')
        
        # Scale
        text_file.write(f'    scale {d["scale"]}\n')
        
        # Transform
        if d["transform"] != "normal":
            text_file.write(f'    transform "{d["transform"]}"\n')
        
        # Position
        text_file.write(f'    position x={d["x"]} y={d["y"]}\n')
        
        # Variable refresh rate (adaptive sync)
        if d.get("adaptive_sync", False):
            text_file.write(f'    variable-refresh-rate\n')
        
        text_file.write(f'}}\n\n')
    
    text_file.close()


def ensure_niri_config_include(config_dir, monitors_file):
    """Ensure that config.kdl includes monitor.kdl"""
    config_kdl = os.path.join(config_dir, "config.kdl")
    monitors_rel_path = "monitor.kdl"
    
    # Check if config.kdl exists
    if not os.path.isfile(config_kdl):
        # Create a minimal config.kdl with include
        with open(config_kdl, "w") as f:
            f.write(f'// Include monitor configuration generated by nwg-displays\n')
            f.write(f'include "{monitors_rel_path}"\n')
        eprint(f"[niri] Created {config_kdl} with include directive")
        return
    
    # Read existing config
    content = load_text_file(config_kdl)
    if content is None:
        return
    
    lines = content.splitlines()
    
    # Check if include already exists
    include_pattern = f'include "{monitors_rel_path}"'
    include_pattern_single = f"include '{monitors_rel_path}'"
    
    for line in lines:
        if include_pattern in line or include_pattern_single in line:
            # Already included
            return
    
    # Add include directive at the beginning
    new_content = f'// Include monitor configuration generated by nwg-displays\ninclude "{monitors_rel_path}"\n\n{content}'
    
    with open(config_kdl, "w") as f:
        f.write(new_content)
    
    eprint(f"[niri] Added include directive to {config_kdl}")


def create_empty_file(file_path):
    if not os.path.isfile(file_path):
        with open(file_path, "w") as file:
            pass


def load_text_file(path):
    try:
        with open(path, 'r') as file:
            data = file.read()
            return data
    except Exception as e:
        print(e)
        return None


def load_workspaces(path, use_desc=False):
    result = {}
    try:
        with open(path, 'r') as file:
            data = file.read().splitlines()
            for i in range(len(data)):
                if data[i] and not data[i].startswith("#"):  # skip comments
                    info = data[i].split("workspace ")[1].split()
                    num = int(info[0])
                    if not use_desc:
                        result[num] = info[2]
                    else:
                        result[num] = data[i].split("output ")[1][1:-1]
            return result
    except Exception as e:
        print(e)
        return result


# We will read all the meaningful lines if -n argument not given or >= number of lines.
def load_workspaces_hypr(path, num_ws=0):
    ws_binds = {}
    meaningful_lines_read = 0
    try:
        with open(path, 'r') as file:
            data = file.read().splitlines()
            r = len(data)
            for i in range(r):
                line = data[i]
                if line and not line.startswith("#"):  # skip comments
                    meaningful_lines_read += 1
                    # Binding workspaces to a monitor, e.g.:
                    # 'workspace=1,monitor:desc:AOC 2475WR F17H4QA000449' or
                    # 'workspace=1,monitor:HDMI-A-1'
                    ws_num = None
                    parts = line.split(",")
                    try:
                        ws_num = int(parts[0].split("=")[1])
                    except:
                        pass
                    mon = parts[1].split(":")[-1]
                    if ws_num:
                        ws_binds[ws_num] = mon

                    if num_ws > 0:
                        if meaningful_lines_read == num_ws:
                            break

            return ws_binds

    except Exception as e:
        eprint("Error parsing workspaces.conf file: {}".format(e))
        return {}


def save_workspaces(data_dict, path, use_desc=False):
    text_file = open(path, "w")
    now = datetime.datetime.now()
    line = "# Generated by nwg-displays on {} at {}. Do not edit manually.\n".format(
        datetime.datetime.strftime(now, '%Y-%m-%d'),
        datetime.datetime.strftime(now, '%H:%M:%S'))
    text_file.write(line + "\n")
    for key in data_dict:
        if not use_desc:
            line = "workspace {} output {}".format(key, data_dict[key])
        else:
            line = "workspace {} output '{}'".format(key, data_dict[key])
        text_file.write(line + "\n")
    text_file.close()


def notify(summary, body, timeout=3000):
    cmd = "notify-send '{}' '{}' -i /usr/share/pixmaps/nwg-displays.svg -t {}".format(summary, body, timeout)
    subprocess.call(cmd, shell=True)


def get_shell_data_dir():
    data_dir = ""
    home = os.getenv("HOME")
    xdg_data_home = os.getenv("XDG_DATA_HOME")

    if xdg_data_home:
        data_dir = os.path.join(xdg_data_home, "nwg-shell/")
    else:
        if home:
            data_dir = os.path.join(home, ".local/share/nwg-shell/")

    return data_dir


def load_shell_data():
    shell_data_file = os.path.join(get_shell_data_dir(), "data")
    shell_data = load_json(shell_data_file) if os.path.isfile(shell_data_file) else {}

    defaults = {
        "interface-locale": ""
    }

    for key in defaults:
        if key not in shell_data:
            shell_data[key] = defaults[key]

    return shell_data
